@@ -1,7 +1,7 @@
 from .core import Primitive, synchronized, PyThread, gated, Timeout, Timer
 from .dequeue import DEQueue
-from .task_queue import TaskQueue, Task
-from .Scheduler import Scheduler, schedule_job, schedule_task, unschedule_job, unschedule_task, global_scheduler
+from .task_queue import TaskQueue, Task, schedule_task
+from .Scheduler import Scheduler
 from .Subprocess import Subprocess, ShellCommand
 from .RWLock import RWLock
 from .Version import Version
@@ -14,8 +14,7 @@ from .producer import Producer
 from .escrow import Escrow
 
 __version__ = Version
-version_info = tuple(Version.split("."))
-
+version_info = (tuple(int(x) for x in Version.split(".")) + (0,0,0))[:3]            # pad with 0s
 
 __a_ll__ = [
     'Primitive',
