@@ -15,6 +15,40 @@ class _WorkerTask(Task):
     def run(self):
         out = self.Processor._process(self.Item, self.Promise)
         
+class _CloseOutputThread(Task):
+    
+    def __init__(self, processor):
+        Thread.__init__(self)
+        self.Processor = processor
+        
+    def close_processor(self):
+        self.Processor.join()
+        self.Processor.close_output()
+            
+    def run(self):
+        pass
+        
+
+class _OutputIterator(object):
+    
+    def __init__(self, processor, queue):
+        self.Queue = queue
+        self.Processor = processor
+        
+    def __next__(self):
+        try:    out, exc_info = self.Queue.pop()
+        except StopIteration:
+            self.Processor._remove_output_queue(self.Queue)
+            raise
+        if exc_info:
+            _, e, _ = exc_info
+            raise e
+        else:
+            return out
+            
+    def __del__(self):
+        self.Processor._remove_output_queue(self.Queue)
+        
 class Processor(Primitive):
     
     Default = ""
@@ -45,7 +79,9 @@ class Processor(Primitive):
         t.start()
     
     def wait_and_close_output(self):
+        print("wait_and_close_output: joining...")
         self.join()
+        print("          join done")
         self.Output.close()
 
     @synchronized
@@ -71,8 +107,10 @@ class Processor(Primitive):
             exc_type, exc_value, tb = sys.exc_info()
             if self.Delegate is not None:
                 self.Delegate.itemFailed(item, exc_type, exc_value, tb)
-            promise.exception(exc_type, exc_value, tb)
             raise
+            promise.exception(exc_type, exc_value, tb)
+            if self.OutputQueue is not None:
+                self.OutputQueue.append((None, (exc_type, exc_value, tb)))
         else:
             #print("%x: out: %s" % (id(self), out))
             if self.Delegate is not None:
@@ -107,8 +145,7 @@ class Processor(Primitive):
         
     def counts(self):
         return self.WorkerQueue.counts()
-
+        
+    @synchronized
     def __iter__(self):
-        if self.Output is None:
-            raise RuntimeError("The processor does not have any output")
         return iter(self.Output)
